#include "reader.hpp"

#include "shaperMap.hpp"

#include <gflags/gflags.h>

DEFINE_uint64(min_part_size, 2048,
    "size of block sent via a single socket that will never be split");

namespace rtransfer {

Reader::Reader(StoragesMap &storages, HandleCache &handleCache,
    std::shared_ptr<ServerSideLink> serverLink)
    : storages_{storages}
    , handleCache_{handleCache}
    , serverLink_{std::move(serverLink)}
{
}

folly::Future<std::size_t> Reader::read(std::uint64_t reqId,
    folly::fbstring storageId, folly::fbstring fileId, folly::fbstring fileGuid,
    std::uint64_t offset, std::size_t size, std::uint8_t priority)
{
    auto it = storages_.find(storageId);
    if (it == storages_.cend())
        return folly::makeFuture<std::size_t>(
            folly::make_exception_wrapper<std::runtime_error>(
                "invalid storage"));

    VLOG(2) << "Reading " << reqId << " offset: " << offset
            << ", size: " << size;
    return handleCache_
        .read(*it->second, reqId, fileId, fileGuid, offset, size, priority)
        .via(folly::getCPUExecutor().get())
        .then([this, reqId, priority, offset](folly::IOBufQueue buf) {
            std::size_t size = buf.chainLength();

            static thread_local std::mt19937 rng{folly::randomNumberSeed()};
            auto dataConns = serverLink_->dataConnections();
            std::shuffle(dataConns.begin(), dataConns.end(), rng);

            std::size_t maxParts = std::min<std::size_t>(
                (size + FLAGS_min_part_size - 1) / FLAGS_min_part_size,
                (1 << 7) - 1);
            std::size_t parts = std::min(maxParts, dataConns.size());
            std::size_t partSize = (size + parts - 1) / parts;

            auto metric =
                "link." + serverLink_->getPeerAddr() + ".pre_write_buffer.size";
            ONE_METRIC_COUNTER_ADD(metric, size);

            VLOG(2) << "Sending " << reqId << ", offset: " << offset
                    << ", size: " << size;

            auto start = std::chrono::steady_clock::now();
            std::vector<folly::Future<folly::Unit>> writeFutures;
            for (auto i = 0u; i < parts; ++i) {
                auto msg = constructDataMsg(
                    reqId, offset, i, parts, buf.splitAtMost(partSize));
                writeFutures.emplace_back(dataConns[i]->write(std::move(msg)));
            }

            return folly::collect(std::move(writeFutures))
                .via(folly::getCPUExecutor().get())
                .then([=, metric = std::move(metric)](
                          const std::vector<folly::Unit> &) {
                    auto now = std::chrono::steady_clock::now();
                    auto delay = now - start;
                    ONE_METRIC_COUNTER_SUB(metric, size);
                    VLOG(2) << "Sent " << reqId << ", offset: " << offset
                            << ", size: " << size << ", took: "
                            << std::chrono::duration_cast<
                                   std::chrono::milliseconds>(delay)
                                   .count()
                            << "ms";

                    return size;
                });
        });
}

std::unique_ptr<folly::IOBuf> Reader::constructDataMsg(std::uint64_t reqId,
    std::uint64_t offset, std::uint8_t part, std::uint8_t totalParts,
    std::unique_ptr<folly::IOBuf> buf)
{
    DCHECK(totalParts < (1 << 7));

    auto nativeSize = buf->computeChainDataLength();

    std::uint32_t size = folly::Endian::big<std::uint32_t>(nativeSize);
    reqId = folly::Endian::big<std::uint64_t>(reqId);
    offset = folly::Endian::big<std::uint64_t>(offset);

    if (part == totalParts - 1)
        part |= proto::last_part_mask;

    auto msg = folly::IOBuf::createCombined(
        sizeof(reqId) + sizeof(part) + sizeof(offset) + sizeof(size));

    auto append = [&](auto val) {
        std::memcpy(msg->writableTail(), &val, sizeof(val));
        msg->append(sizeof(val));
    };

    append(reqId);
    append(part);
    append(offset);
    append(size);

    if (nativeSize > 0)
        msg->prependChain(std::move(buf));

    return msg;
}

}  // namespace rtransfer
